﻿using System;
using System.Diagnostics;
using glb;


namespace evc
{
	// リソース状態
	public enum siResourceState
	{
		Unused = 0,		// 未使用
		Loading,		// ロード中
		Initializing,	// 初期化中
		Effective,		// 使用可能リソース
		Invalidity,		// 無効リソース
	};

	// リソース初期化戻り値
	public enum siResourceInitType
	{
		Success,		// 成功
		Failure,		// 失敗
		Continuation,	// 続行
	};


	// １リソース情報
	// このクラスはリソースの共通部分です。このクラスを継承した
	// siResourceSound,siResourceTextureに対して処理を実装して下さい。
	// 実装が必要なメソッドは
	// 
	// bool Init( byte[] abData );
	// void Dispose();
	// 
	// のみです。
	public abstract class siResourceBase : IDisposable
	{
		private siFile	inFile = null;

		// コンストラクタ
		protected siResourceBase ( string strName )
		{
			this.Name = strName;
			this.State = siResourceState.Unused;	// リソースの状態
		}

		// リソース名(ファイル名)取得・設定
		public string Name					{ get; private set; }
		// リソースの状態を取得
		public siResourceState State		{ get; protected set; }
		// シーン読み込み時にロードするデータか否か
		public bool IsAutoLoad				{ get; set; }


		// リソースタイプ
		public abstract seResourceType Type	{ get; }

		// 初期化が終了したか否か
		public bool IsEffective				{ get { return (this.State == siResourceState.Effective); } }



		// リソースロード(名前は事前指定)
		public virtual void Load( seSceneManager inSceneManager )
		{
			// 既に初期化済なら無処理
			if ( this.IsEffective ) return;

			switch (this.State)
			{
				// 現在初期化中なら
				case siResourceState.Loading:
				case siResourceState.Initializing:
				case siResourceState.Effective:
					break;
				
				// データが全くない状態なのでロード
				case siResourceState.Unused:
				case siResourceState.Invalidity:
					Debug.Assert( this.inFile == null );
					siIOManager inIOManager = seRootScene.Instance.IOManager;
					this.inFile = inIOManager.CreateLoader( inSceneManager, this.Name );
					this.State = siResourceState.Loading;
					break;
				default:
					break;
			}

		}

		// リソース状態更新
		public virtual void Update()
		{
			switch (this.State)
			{
				case siResourceState.Unused:
					break;
				// ロード中
				case siResourceState.Loading:
					Debug.Assert( this.inFile != null );

					// ロード状態更新。
					this.inFile.Update();
					// ロード終了していない
					if ( !this.inFile.IsLoadFinish ) break;
					// 初期化中に移行
					this.State = siResourceState.Initializing;
					break;

				// 初期化中
				case siResourceState.Initializing:

					// 初期化終了処理
					glbAction<siResourceState> inEndFile = delegate( siResourceState eState )
					{
						// ローダーを解放
						this.inFile.Dispose();
						this.inFile = null;
						this.State = eState;
					};

					switch (this.Init( this.inFile.Buffer ))
					{
						// 初期化成功
						case siResourceInitType.Success: inEndFile(siResourceState.Effective); break;
						// 初期化失敗
						case siResourceInitType.Failure: inEndFile(siResourceState.Invalidity); break;
						// 初期化継続
						case siResourceInitType.Continuation: break;
					}
					break;

				case siResourceState.Effective:
					break;
				case siResourceState.Invalidity:
					break;
			}
		}

		// リソースのキャンセル
		internal virtual void Cancel()
		{
			// ローダー終了
			if ( this.inFile != null )
			{
				this.inFile.Dispose();
				this.inFile = null;
			}
		
			// 初期化済みなら
			if ( this.IsEffective )
			{
				this.Dispose();
			}
			this.State = siResourceState.Unused;
		}

		// ロード済みのデータに対して初期化を行う
		public void Register( byte[] abData )
		{
			if ( this.IsEffective ) return;		// 既に初期化済み

			switch (this.Init( abData ))
			{
				case siResourceInitType.Success: this.State = siResourceState.Effective; break;
				case siResourceInitType.Failure: this.State = siResourceState.Invalidity; break;
				case siResourceInitType.Continuation: this.State = siResourceState.Initializing; break;
			}
		}



		// 初期化処理:リソースの初期化処理を実装して下さい。
		// 呼び出されるタイミングはリソースの種類別に以下の通りです。
		//
		// ・テクスチャ(siResourceTexture)
		// リソースデータのロード終了時に呼び出されます。
		//
		// ・サウンド(siResourceSound)
		// リソースが要求された時点で呼び出されます。データはロードされず、
		// abDataにはnullが入ります。
		//
		// Init時、ファイル名を取得する場合はthis.Nameを使用して下さい。
		//
		// @retval true:初期化完了 false:初期化中
		public abstract siResourceInitType Init( byte[] abData );

		// リソース解放：リソースを使い終わった時の処理を実装して下さい。
		public abstract void Dispose();
	}


	// サウンドリソース
	public abstract class siResourceSound : siResourceBase
	{
		protected siResourceSound ( string strSoundName ) : base( strSoundName ){}
		// リソースタイプ
		public override seResourceType Type	{get{return seResourceType.Sound;}}
	}

	// テクスチャリソース
	public abstract class siResourceTexture : siResourceBase
	{
		protected siResourceTexture ( string strTextureName ) : base( strTextureName ){}

		// イメージの左上を(0,0)右下を(TextureSize.X,TextureSize.Y)とした時、
		// 座標(iX, iY)がイメージとヒットしているかを返してください。
		// ドット単位で当たり判定をする時に使用します。
		// 判定をしない場合はtrueを返してください。
		public abstract bool IsHit( glbVector2I sPos );
	
		// テクスチャサイズ:
		// テクスチャの幅、高さを返す処理を実装して下さい。
		public abstract glbVector2I TextureSize { get; }
		
		// リソースタイプ
		public override seResourceType Type	{get{return seResourceType.Image;}}
	}


}
