﻿using System.Collections.Generic;
using System.Diagnostics;
using glb;
#if SL_PROC
using sl;
#endif

namespace evc
{
	// ＩＤを管理するクラス
	// 重複しないようにＩＤを生成、割り当てます。

	public partial class seIDManager<T>
		where T : class
	{
		public const int InvalidID = -1;	// 無効ID
		public const int DefaultBaseID = 1000; // 未設定時のデータ開始ＩＤ

		private SortedDictionary<int, T>	dicElemObjectList = new SortedDictionary<int,T>();

		// ベースＩＤ(最低値)の取得・設定
		public int BaseID { get; private set; }
		// オブジェクト数取得
		public int Count { get { return this.dicElemObjectList.Count; }}


		// コンストラクタ
		public seIDManager() : this( seIDManager<T>.DefaultBaseID ){}
		public seIDManager( int iInBaseID )
		{
			this.BaseID = iInBaseID;
		}


		// 内部データをクリア
		public virtual void Clear()
		{
			this.dicElemObjectList.Clear();
		}

		// 任意のデータを保持しているか判定
		public bool ContainsValue( T tData )
		{
			return this.dicElemObjectList.ContainsValue( tData );
		}


		// 全てのデータに対して処理を実行
		public bool DoAll( glbFunc<int,T, bool> func )
		{
			foreach( KeyValuePair<int,T> pair in this.dicElemObjectList )
			{
				if (!func(pair.Key, pair.Value)) return false;
			}
			return true;
		}

		// 全てのデータに対して処理を実行(逆順に処理)
		public bool DoAllReverse(glbFunc<int,T, bool> func)
		{
			// 再帰的にデータを引き出していき、最後で引き返す。
			glbFunc<SortedDictionary<int, T>.Enumerator, bool> revFunc = null;
			revFunc = delegate( SortedDictionary<int, T>.Enumerator ItNow )
			{
				KeyValuePair<int,T> pair = ItNow.Current;
				
				// 次があるなら再帰的に呼び出す
				if (ItNow.MoveNext())
				{
					if (!revFunc(ItNow)) return false;
				}

				return func(pair.Key, pair.Value);
			};
	
			SortedDictionary<int, T>.Enumerator It = this.dicElemObjectList.GetEnumerator();
			if ( !It.MoveNext() ) return true;
			return revFunc( It );
		}



		// データを削除
		public virtual T Remove( int iIndex )
		{
			T tData = this.Find( iIndex );
			// 要素オブジェクトを検索
			this.dicElemObjectList.Remove( iIndex );

			// 削除したインスタンスを返す。
			return tData;
		}

		// データを削除(要素指定)
		public virtual int Remove( T tData )
		{
			int iRemoveID = seIDManager<T>.InvalidID;
			this.DoAll( delegate( int iID, T tValue )
			{
				if ( tValue.Equals( tData ) )
				{
					iRemoveID = iID;
					return false;
				}
				return true;
			});
			if (iRemoveID != seIDManager<T>.InvalidID)
			{
				this.dicElemObjectList.Remove( iRemoveID );
			}
			return iRemoveID;
		}


		// データを追加
		public virtual void Add( int iIndex, T tData )
		{
			// 無いのがあたりまえ。
			Debug.Assert( !this.dicElemObjectList.ContainsKey( iIndex ) );
			this.dicElemObjectList[ iIndex ] = tData;
		}
		public virtual int Add( T tData )
		{
			int iIndex = this.GetFreeID();
			// ここでthis.Add(int,T)を呼ぶと、下位でAdd(T)、Add(int,T)両方を
			// オーバーライドしていた場合、ユーザの意図しない結果になる可能性があるので
			// あえて同じ処理を書く。
			Debug.Assert( !this.dicElemObjectList.ContainsKey( iIndex ) );
			this.dicElemObjectList[ iIndex ] = tData;
			return iIndex;
		}

		// 要素オブジェクトを検索
		public T Find( int iIndex )
		{
			if ( !this.dicElemObjectList.ContainsKey( iIndex ) ) return null;
			return this.dicElemObjectList[ iIndex ];
		}

		// データからキーIDを取得
		public int Find( T tData )
		{
			int iRet = seIDManager<T>.InvalidID;
			this.DoAll( delegate( int iID, T tInData )
			{
				if ( tInData.Equals( tData ) )
				{
					iRet = iID;
					return false;
				}
				return true;
			});
			return iRet;
		}


		// 次のＩＤを取得(dicIndexに登録されていないKeyを返す)
		// 整合性を保つため、何度呼ばれても同じ値を返すアルゴリズムにする。
		public int GetFreeID()
		{
			int iCount = this.dicElemObjectList.Count;

			if ( iCount == 0 )
			{
				return this.BaseID;
			}

			int iBfoID = this.BaseID;
			foreach ( int iIndex in this.dicElemObjectList.Keys )
			{
				if ( ( iIndex - iBfoID ) > 1 )
				{
					return iBfoID + 1;
				}
				iBfoID = iIndex;
			}
	
			return iBfoID + 1;
		}
	}

}
