﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Diagnostics;
using evc;
using glb;


namespace evc
{
	// ロード用イベント命令タイプ
	internal enum LoadCmdType : int
	{
		FlowIf = 0,		// if
		Cond = 2,		// 実行条件
		Event = 3,		// 通常のイベント命令
		FlowNon = 4,	// 無効フロー(if文のネスト判定時、次がない場合に指定)
		Call = 5,		// サブルーチンコール
	}
	

	// 実行条件管理
	public partial class evCmdFlowCondManager
	{
		// 実行条件
		private List<evCmdCond> lisCondCmd = new List<evCmdCond>();
		public ReadOnlyCollection<evCmdCond> Cond { get { return this.lisCondCmd.AsReadOnly(); } }

		public evCmdFlowCondManager()
		{
		}

		public evCmdFlowCondManager( evCmdFlowCondManager mst )
		{
			foreach (var cond in this.lisCondCmd)
			{
				this.Add(new evCmdCond(cond));
			}
		}

		// 命令の追加
		public void Add( evCmdCond inCmd )
		{
			this.lisCondCmd.Add( inCmd );
		}

		// 命令の削除
		public void Remove( evCmdCond inCmd )
		{
			this.lisCondCmd.Remove( inCmd );
		}

		// 実行条件を満たしているか判定
		// true:満たしている  false:満たしていない
		public bool IsCondAll( IevCmdUI inCmdUI )
		{
			// 条件が無い場合は無条件で合格
			if ( this.lisCondCmd.Count <= 0 ) return true;

			foreach (evCmd inCond in this.lisCondCmd )
			{
				// 有効で無い場合はスルーなので合格
				if ( !inCond.Enable ) continue;
				if ( !inCond.DoFunc( inCmdUI ) ) return false;
			}
			return true;
		}

		// データロード
		public void Load( IseStateInput inReader, IevCmdUI inCmdUI )
		{
			this.lisCondCmd.Clear();
			int iCount = inReader.ReadShort();
			seDefine.DebugTrace( "CondCount", iCount );
			for (int i = 0; i < iCount; i++)
			{
				evCond eCond = (evCond)inReader.ReadByte();
				seDefine.DebugTrace( eCond.ToString(), 0 );
				evCmdCond inCmdCond = evCmdList.CraeteCmd( eCond, inCmdUI );
				inCmdCond.Load( inReader, inCmdUI );
				this.Add( inCmdCond );
			}
		}
	}


	// if
	public partial class evCmdFlowIf : evCmdHeader
	{
		// 実行条件
		public evCmdFlowCondManager CondManager { get; private set; }
		// イベント
		private List<evCmdHeader> lisCmd = new List<evCmdHeader>();
		public ReadOnlyCollection<evCmdHeader> Cmd { get { return this.lisCmd.AsReadOnly(); } }

		// フローの１つ前のチェーン
		public evCmdFlowIf BefoFlow { get; private set; }

		// フローのチェーン
		private evCmdFlowIf inNextFlow = null;
		public evCmdFlowIf NextFlow 
		{
			get { return this.inNextFlow; } 
			set 
			{
				if ( value != null )
				{
					value.Parent = this.Parent;
					value.BefoFlow = this;
				}
				this.inNextFlow = value; 
			}
		}


		// コンストラクタ
		public evCmdFlowIf()
		{
			this.NextFlow = null;
			this.CondManager = new evCmdFlowCondManager();
		}

		// コンストラクタ(クローン用)
		public evCmdFlowIf( evCmdFlowIf mst )
		{
			this.NextFlow = (mst.NextFlow == null) ? null : new evCmdFlowIf(mst.NextFlow);
			this.CondManager = new evCmdFlowCondManager(mst.CondManager);

			foreach (var cmdHeader in mst.Cmd)
			{
				this.Add(cmdHeader.Clone());
			}
		}


		// 命令の追加
		public void Add( evCmdHeader inCmd )
		{
			this.lisCmd.Add( inCmd );
			// フローなら親を設定
			inCmd.Parent = this;
		}

		// 命令の削除
		public void Remove( evCmdHeader inCmd )
		{
			this.lisCmd.Remove( inCmd );
			// 親を削除
			inCmd.Parent = null;
		}


		// 自分以下の全ての命令に対して処理を実行
		public bool DoAll( glbFunc<evCmdHeader,bool> func )
		{
			if ( !func( this ) ) return false;

			// 条件実行
			foreach (var cond in this.CondManager.Cond)
			{
				if( !func(cond) ) return false;
			}

			// イベントに対して処理を実行
			foreach (var header in this.lisCmd)
			{
				// 条件フローなら
				evCmdFlowIf f = header as evCmdFlowIf;
				if ( f != null )
				{
					if (!f.DoAll(func)) return false;
				}
				// 通常のイベント命令なら
				else
				{
					if (!func(header)) return false;
				}
			}

			// 次の連結フローがあるならそれを実行
			if ( this.NextFlow != null)
			{
				return this.NextFlow.DoAll( func );
			}

			return true;
		}


		// イベントの実行
		// true:実行した　false:実行しなかった
		public void DoEvent( IevCmdUI inCmdUI )
		{
			// 処理を実行
			this.lisCmd.ForEach( delegate( evCmdHeader inEvent )
			{
				if ( inEvent.Enable )
				{
					inEvent.DoFunc( inCmdUI );
				}
			});
		}


		// クローン：再生時専用です。
		// 編集で使ういくつかのデータはコピーされていません。
		public override evCmdHeader Clone()
		{
			return new evCmdFlowIf(this);
		}


		// 
		public override bool DoFunc( IevCmdUI inCmdUI )
		{
			// フローが有効でかつ、条件を満たしたなら
			if ( this.Enable && this.CondManager.IsCondAll( inCmdUI ) )
			{
				this.DoEvent( inCmdUI );
			}
			else if ( this.NextFlow != null )
			{
				this.NextFlow.DoFunc( inCmdUI );
			}
			return true;
		}

		// データロード
		public override void Load(IseStateInput inReader, IevCmdUI inCmdUI)
		{
			int iCount = (int)inReader.ReadShort();
			seDefine.DebugTrace( "CmdCount", iCount );
			for (int i = 0; i < iCount; i++)
			{
				// タイプ別処理
				evCmdHeader inHeader = null;
				LoadCmdType eCmdType = (LoadCmdType)inReader.ReadByte();
				seDefine.DebugTrace( eCmdType.ToString(), 0 );
				switch (eCmdType)
				{
					case LoadCmdType.FlowIf:
						inHeader = new evCmdFlowIf();
						break;
					case LoadCmdType.Event:
						evEvent eEvent = (evEvent)inReader.ReadByte();
						inHeader = evCmdList.CraeteCmd( eEvent, inCmdUI );
						break;
					default:
						Debug.Assert( false );
						break;
				}

				inHeader.Load( inReader, inCmdUI );
				this.Add( inHeader );
			}

			this.CondManager.Load( inReader, inCmdUI );

			LoadCmdType eNextType = (LoadCmdType)inReader.ReadByte();
			evCmdFlowIf inFlow = null;
			switch (eNextType)
			{
				case LoadCmdType.FlowIf:
					inFlow = new evCmdFlowIf();
					break;
				case LoadCmdType.FlowNon:
					break;
				default:
					Debug.Assert( false );
					break;
			}

			// 親をコピー
			this.NextFlow = inFlow;
			if ( inFlow != null )
			{
				inFlow.Load( inReader, inCmdUI );
			}
		}
	}

}
