﻿using System.Collections.Generic;
using glb;

namespace evc
{
	// エレメントのＩＤとrootの管理をします。
	// 主にフレーム、カメラの管理をします。
	public partial class seElemManager : seIDManager<seElem>
	{
		private seSceneManager _scene = null; // 操作対象シーン
		private float fPlayTime = 0.0f;			// 現在の再生時間
		private int	iFrame = 0;					// 現在のフレーム(主に編集用)

		// 描画の最初と最後のエレメント
		private seElem _firstDrawElem = null;
		private seElem _lastDrawElem = null;

		internal int LastAnim { get; private set; }		// 最後に要求されたアニメーション


		// ルート要素取得・設定
		private seElem _rootElem = null;			// 最上位要素
		public seElem RootElem 
		{ 
			get { return this._rootElem; } 
			set 
			{ 
				this._rootElem = value;
				this._rootElem.State.ClearStateData();
				seStateFrameCtrl fc = (seStateFrameCtrl)seStateUtil.Instance.CreateStateType(emElemState.FrameCtrl);
				fc.SceneManager = this._scene;
				this._rootElem.State.LinkStateData(emElemState.FrameCtrl, fc );
			}
		}

		// コンストラクタ
		public seElemManager ( seSceneManager scene )
		{
			this._scene = scene;
			this.LastAnim = seAnimManager.InvalidID;
		}


		// 優先順位管理
		private class DrawPriority
		{
			internal seElem _lastElem = null;
			internal seElem _firstElem = null;

			// クリア
			internal void Clear()
			{
				this._firstElem = null;
				this._lastElem = null;
			}

			// 登録
			internal void Add( seElem elem )
			{
				if ( this._firstElem == null )
				{
					this._firstElem = elem;
					this._lastElem = elem;
					this._firstElem.BeforDraw = null;
				}
				else
				{
					this._lastElem.NextDraw = elem;
					elem.BeforDraw = this._lastElem;
					this._lastElem = elem;
				}
				this._lastElem.NextDraw = null;
			}

		}


		// 描画順に表示する全てのエレメントに対して処理を実行
		public bool DoDrawAll( glbFunc<seElem,bool> func )
		{
			seElem elem = this._firstDrawElem;
			while (elem != null)
			{
				if ( !func( elem ) ) return false;
				elem = elem.NextDraw;
			}
			return true;
		}

		// 描画リストワーク:一時データです。セーブの必要はありません。
		// 毎回作ることも可能ですが、処理速度を期待してメンバとして持っておきます。
		private SortedDictionary<int,DrawPriority> dicDrawList = new SortedDictionary<int,DrawPriority>();

		// 描画エレメントリストを作成
		private void RegisterDrawList()
		{
			glbAction<seElem> RegisterElem = null;
			RegisterElem = delegate( seElem elem )
			{
				// 編集OFFフラグが立っているならそのエレメント以下は無処理
				if (elem.DisableEditParent) return;

				// 表示優先度を取得

				// 描画可能なら
				if ( elem.IsVisible )
				{
					int iPriority = elem.State.NowPriority;
					DrawPriority drawWork = null;
					if ( !this.dicDrawList.TryGetValue( iPriority, out drawWork ) )
					{
						drawWork = new DrawPriority();
						this.dicDrawList.Add( iPriority, drawWork );
					}
					drawWork.Add( elem );
				}

				// 子に対して登録を続行
				foreach (var childElem in elem.Child)
				{
					RegisterElem( childElem );
				}
			};

			// 最後に描画した要素初期化
			this.dicDrawList.Clear();
			this._firstDrawElem = null;
			this._lastDrawElem = null;

			if ( this._rootElem != null ) 
			{
				// 登録開始
				RegisterElem( this._rootElem );

				// プライオリティ毎のエレメントを連結して
				// １次配列の描画リストを作成する
				DrawPriority bfoWork = null;
				foreach (DrawPriority drawWork in this.dicDrawList.Values)
				{
					if ( bfoWork != null )
					{
						bfoWork._lastElem.NextDraw = drawWork._firstElem;
						drawWork._firstElem.BeforDraw = bfoWork._lastElem;
					}
					else
					{
						this._firstDrawElem = drawWork._firstElem;
						this._firstDrawElem.BeforDraw = null;
					}

					this._lastDrawElem = drawWork._lastElem;
					bfoWork = drawWork;
				}

			}
		}



#if true
		// 全体に影響を与えるカラーを取得・設定
		public glbColor Color
		{
			get
			{
				if (this._rootElem == null) return glbColor.White;
				return this._rootElem.State.NowColor;
			}
			set
			{
				if (this._rootElem == null) return;
				this._rootElem.State.NowColor = value;
			}
		}

#endif



		// PlayTime = FrameTime * Frame
		// この３つの値は常に同期させる。

		// プレイ時間取得・設定
		public float PlayTime
		{
			get { return this.fPlayTime; }
			set
			{
				this.fPlayTime = value;
				float fFrameTime = this._scene.Option.FrameTime;
				this.iFrame = (int)( this.fPlayTime / fFrameTime );
			}
		}
		// フレームの取得・設定
		public int Frame
		{
			get { return this.iFrame; }
			// 外部からはイベントを通してのみ設定可能です。
			set
			{
				this.iFrame = value;
				float fFrameTime = this._scene.Option.FrameTime;
				this.fPlayTime = fFrameTime * (float)this.iFrame;
			}
		}

		// クリア
		internal new void Clear()
		{
			base.Clear();

			// ルート以下を削除
			this._rootElem = null;
			this.fPlayTime = 0.0f;	
			this.iFrame = 0;
		}


		// 全てのエレメントに対して処理を実行
		public bool DoAll( glbFunc<seElem,bool> inFunc )
		{
			if ( this._rootElem == null ) return false;
			return this._rootElem.DoAll( inFunc );
		}



		// アニメーション設定
		public void SetAnim( int iAnimID )
		{
			this.LastAnim = iAnimID;

			// 特殊処理：FrameCtrlだけはアニメーション毎に初期化(状態を継承しない）
			// FrameCtrlはrootのみ存在する
			this._rootElem.State.GetData( emElemState.FrameCtrl ).ChangeAnim( null );

			seRootScene rs = seRootScene.Instance;

			this.DoAll( delegate( seElem elem )
			{
				// 編集モードの時はアニメーションを一度クリア
				if ( rs.EditMode )
				{
					elem.State.ClearAnim();
				}

				// アニメーションを検索して存在した時のみ処理を実行
				seElemAnim elemAnim = elem.Anim.Find(iAnimID);
				if ( elemAnim == null ) return true;

				elemAnim.DoAll( delegate ( seStateKeyManager keyManager )
				{
					seStateData state = elem.State.GetData( keyManager.StateType );
					state.ChangeAnim( keyManager );
				});

				return true;
			});

			// 存在するならアニメーションリスナーへ通知
			if ( rs.AnimListener != null )
			{
				seAnimInfo ai = rs.RootScene.AnimManager.Find( iAnimID );
				rs.AnimListener.OnAnimStart( ai.Name );
			}
		}


		// ルート以下のアニメーションを再更新
		public void ResetAnim()
		{
			if (this._rootElem == null) return;
			float fFrameTime = this._scene.Option.FrameTime;

			this.ResetAnim(this.Frame * fFrameTime, fFrameTime, seRootScene.Instance.Camera);
		}
		// ルート以下のアニメーションを再更新
		public void ResetAnim(float fTime, float fFrameTime, IGLMatrix inCamera)
		{
			this.DoAll(delegate(seElem elem)
			{
				elem.ResetAnim(inCamera, fFrameTime, fTime);
				return true;
			});

			// 描画リスト更新
			this.RegisterDrawList();
		}

		public void Draw()
		{
			siMaterial material = seRootScene.Instance.Material;
			material.BlendType = evc.BlendType.NORMAL;
			material.DoublyType = evc.DoublyType.Normal;
			material.MaskAlpha = 255;
			material.Apply();

			// 登録済みのエレメントに対して描画
			this.DoDrawAll( delegate( seElem elem )
			{
				elem.Draw( material );
				return true;
			});

			// 最後にマスクのクリアをしておく。
			seRootScene.Instance.IOManager.ClearMask();
		}

		// 更新
		internal void Update( float fAddTime )
		{
			this.Update( seRootScene.Instance.Camera, fAddTime );
		}

		// 更新:(カメラ指定)
		internal void Update( IGLMatrix mParent, float fAddTime )
		{
			// 要素アニメーション実行
			if ( this._rootElem != null )
			{
				this.PlayTime += fAddTime;

				float fFrameTime = this._scene.Option.FrameTime;

				// 再帰的にエレメントを見ていきアニメーションを更新
				glbAction<seElem> UpdateElem = null;
				UpdateElem = delegate( seElem elem )
				{
					// falseなら下位のエレメントの処理はしない
					if ( !elem.Update( mParent, fFrameTime, fAddTime ) )
						return;

					foreach (var childElem in elem.Child)
					{
						UpdateElem( childElem );
					}
				};
				UpdateElem( this._rootElem );

			}
			// 描画リスト更新
			this.RegisterDrawList();
		}


		// 親子継承を更新
		public void Succession( IGLMatrix mCamera )
		{
			this.DoAll( delegate( seElem elem )
			{
				elem.Succession( mCamera );
				return true;
			});
		}

		// オブジェクトワーク再設定
		public void ResetObjectWork( seObject newObject, seObject oldObject )
		{
			this.DoAll(delegate(seElem elem)
			{
				if ( elem.ObjectWork == null ) return true;

				// 描画オブジェクトの再割り当て。
				if (elem.ObjectWork.Object == oldObject)
				{
					elem.ObjectWork = (newObject == null) ?
							null : newObject.CreateDrawWork( elem );
				}
				return true;
			});
		}


		// イベント関係
		// 

		// ポイントのヒット状態を更新
		// @param bFirstHit : 最初にヒットしたフラグ
		public bool UpdateHit( glbVector2 point, bool bFirstHit )
		{
			// 描画した要素を手前から順に処理していく
			seElem elem = this._lastDrawElem;
			for (; elem != null; elem = elem.BeforDraw)
			{
				// イベント用フラグをクリア
				elem.Hit = seHit.Non;

				// シーン要素ならシーン内の当たり判定
				seoSceneWork sceneWork = elem.ObjectWork as seoSceneWork;
				if (sceneWork != null)
				{
					// シーンがロードされていない場合はnull
					seSceneManager localScene = sceneWork.LocalScene;
					if (localScene != null)
					{
						bFirstHit = localScene.ElemManager.UpdateHit( point, bFirstHit );
					}
				}
				// それ以外は通常の当たり判定
				else if (elem.IsInRect( point ))
				{
					bool bHit = true;
					// テクスチャでピクセル単位あたり判定なら
					seStateTextureControl inTexControl = (seStateTextureControl)elem.State[emElemState.TextureControl];
					if ( (inTexControl != null) && (inTexControl.Hit == seTexHit.Pixel) )
					{
						// ピクセル単位で当たり判定
						bHit = elem.IsHitPixel( point );
					}

					if ( bHit )
					{
						elem.Hit = (bFirstHit) ? seHit.TopHit : seHit.Hit;
						bFirstHit = false;
					}
				}
			}
			return bFirstHit;
		}

		// イベント実行処理
		public void DoEvent()
		{
			// イベント全てに対して処理を実行
			// 非表示でも実行
			// 当たり判定は非表示のエレメントは失敗するようになっています。
			// 
			if ( this._rootElem == null ) return;

			this._rootElem.DoAll( delegate( seElem elem )
			{
				if ( elem.Event != null )
				{
					elem.Event.DoFunc( this._scene );
				}
				// シーンなら中のイベントも実行
				seoSceneWork sw = elem.ObjectWork as seoSceneWork;
				if ( sw != null ) 
				{
					sw.DoEvent();
				}
				return true;
			});
		}


		// ロード
		internal void Load( IseStateInput input, seSceneManager scene  )
		{
			this.RootElem = new seElem( string.Empty, null, seElemManager.InvalidID );
			this.RootElem.Load( input, scene );

			// エレメントの登録終了後に実行
			scene.ObjectManager.DoAll( delegate( int objID, seObject obj )
			{
				// メッシュに対してメッシュデータの適応
				if ( obj.ElemType == emObjectType.Mesh )
				{
					// これにより、リンクする全てのエレメントも更新されます。
					seoMesh mesh = (seoMesh)obj;
					mesh.RegisterMesh();
					mesh.UpdateMeshElem();
				}
				return true;
			});
		}

	}
}
