﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using glb;

namespace evc
{
	public class seKeyWord : seStringPair{}

	/**
	 * シーン毎管理
	 */
	public partial class seSceneManager
	{
		public string Name							{ get; set; }			// シーン名取得
		public seKeyWord KeyWord					{ get; private set; }	// キーワード管理を取得
		public seOption Option						{ get; private set; }	// 設定管理を取得
		public seObjectManager ObjectManager		{ get; private set; }	// オブジェクトマネージャーを取得
		public seAnimManager AnimManager			{ get; private set; }	// アニメーションマネージャの取得
		public seResourceManager ResourceManager	{ get; private set; }	// テクスチャコントロール取得
		public seElemManager ElemManager			{ get; private set; }	// エレメントマネージャ取得
		public seKeyControlManager KeyControlManager{ get; private set; }	// コントロールキー管理取得
		public evVariableManager VariableList		{ get; private set; }	// グローバル変数リスト取得・設定

		private bool				bClone;								// 複製したシーンか否か
		private List<int>			lisChangeAnim = new List<int>();	// 次のフレームで切り替えるアニメーション



		#region アニメーションリクエスト関係

		// 次のフレーム実行されるアニメーションか否か
		public bool IsPlayNextAnim(int iAnimID)
		{
			return this.lisChangeAnim.Contains(iAnimID);
		}

		// 次のフレームで切り替えるアニメーション
		// アニメーションは重ねて発生する可能性があるので複数登録できるようにしておく。
		public void RequestAnim(int iAnimID)
		{
			// 無効のアニメーションリクエストは無視する
			if (seAnimManager.InvalidID == iAnimID) return;

			// 無い場合のみ追加
			if (!this.lisChangeAnim.Contains(iAnimID))
			{
				this.lisChangeAnim.Add(iAnimID);
			}
		}

		// 名前を指定してアニメーションリクエスト
		public void RequestAnim(string animName)
		{
			this.RequestAnim( this.AnimManager.FindID(animName) );
		}

		// リクエストされていたアニメーションを呼び出す
		// アニメーションの変更は更新タイミングで１度に行います。
		internal void UpdateAnim()
		{
			// 次のアニメーションが設定されているならそれを実行
			foreach (var animID in this.lisChangeAnim)
			{
				this.ElemManager.SetAnim(animID);
			}
			this.lisChangeAnim.Clear();
		}

		#endregion


		// 親シーンを取得・設定
		private seSceneManager	_parent;
		public seSceneManager Parent
		{
			get { return this._parent; }
			set 
			{ 
				this._parent = value;
				// 親のキーワードとリンク
				this.KeyWord = (this._parent != null) ? this._parent.KeyWord : null;
			}
		}


		// コンストラクタ
		public seSceneManager( string name )
		{
			// 複製ではない。
			this.bClone = false;

			// キーワードファイルはrootにのみ存在します。
			// rootのデータロード時にインスタンスを作成、読み込みます。
			this.KeyWord = null;

			this.Name = name;
			this.Option = new seOption();
			this.ElemManager = new seElemManager( this );
			this.ObjectManager = new seObjectManager( this );
			this.AnimManager = new seAnimManager();
			this.ResourceManager = new seResourceManager();
			this.KeyControlManager = new seKeyControlManager();
			// 変数だけは特殊。クローンの場合はコピーを作る。
			this.VariableList = new evVariableManager();
		}

		// コンストラクタ(Clone用)
		// 編集しないものは元からリンクし、共有する。
		private seSceneManager(seSceneManager sm)
		{
			// 複製シーン。
			this.bClone = true;

			this.Name = sm.Name;
			this.ElemManager = new seElemManager(this); // 引数が必要な為、クローン時もここで作る。
			if ( sm.ElemManager.RootElem != null )
			{
				// データをコピー
				this.ElemManager.RootElem = sm.ElemManager.RootElem.Clone( null, this );
			}
			this.Option				= sm.Option;
			this.ObjectManager		= sm.ObjectManager;
			this.AnimManager		= sm.AnimManager;
			this.ResourceManager	= sm.ResourceManager;
			this.KeyWord			= sm.KeyWord;
			this.KeyControlManager	= sm.KeyControlManager;

			// 変数データのクローンを作成
			this.VariableList = sm.VariableList.Clone();
		}


		// シーンをリセット
		public void Clear()
		{
			// キーワードは全シーンで共有の為削除されません。
			// 

			// リクエスト中のアニメーションをクリア
			this.lisChangeAnim.Clear();

			this.Option.Clear();
			this.ElemManager.Clear();
			this.VariableList.Clear();

			// クローンの時は共有マネージャをクリアしない。
			if (!this.bClone)
			{
				this.ObjectManager.Clear();
				this.AnimManager.Clear();
				this.ResourceManager.Clear();
				this.KeyControlManager.Clear();
			}

			this.Name = string.Empty;
		}


		// シーンの複製を作成
		// テクスチャ、描画オブジェクト、アニメーションは共有、
		// 要素はデータが更新されるのでデータを複製
		public seSceneManager Clone()
		{
			return new seSceneManager( this );
		}


		// シーンプロシージャ
		public void OnProc( float fAddTime )
		{
			this.UpdateAnim();
			// 要素アニメーション実行
			this.ElemManager.Update( fAddTime );
		}


		// バイナリロード
		public void Load( IseStateInput input )
		{
			// マジックが違います。praファイルではありません。
			if ( input.ReadInt() != seDefine.Magic )
			{
				Debug.Assert( false );
				return;
			}

			// バージョンが違います。ファイルと読み込み処理のバージョンは
			// 揃える必要があります。
			if ( input.ReadInt() != seDefine.FileVarsion )
			{
				Debug.Assert( false );
				return;
			}

			// 変数リスト
			this.VariableList.Load( input );

			// オプション
			this.Option.Load( input );

			// データが存在するならキーワードファイルをロード
			if ( this.Option.IncludeKeyWord )
			{
				this.KeyWord = new seKeyWord();
				this.KeyWord.Load( input );
			}
	
			// リソース入力
			this.ResourceManager.Load( input, this );

			// オブジェクト入力
			this.ObjectManager.Load( input );

			// アニメーションリスト入力
			this.AnimManager.Load( input );

			// 存在するなら要素入力
			this.ElemManager.Load( input, this );
		}


	}
}
