﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using glb;


namespace evc
{
	// 当たり判定タイプ
	public enum seHit
	{
		Non = 0,	// ヒットしていない
		TopHit,		// 一番最初にヒット
		Hit			// ヒット
	}

	/**
	 * １要素管理クラス
	 */
	public partial class seElem
	{
		// ↓以下は毎フレーム設定されます。保存されません。
	
		public bool StopAnim	{ get; set; }			// アニメーション停止
		public bool StopAnimAll { get; set; }			// 自分以下のアニメーション停止
		public seHit Hit		{ get; internal set; }	// 当たり判定状態

		// ↑以上は毎フレーム設定されます。保存されません。

		public bool DisableEdit { get; set; }	// 編集を無視(編集用):完全に非表示にします。
		public bool DisableEditParent { get; set; }	// 自分より上に編集無視が存在することを表します。(編集用)

		
		public seElem			 Parent		{ get; private set; }	// 親
		public int				 ElemID		{ get; private set; }	// 識別ＩＤを取得
		public string			 Name		{ get; set;			}	// 名前
		public seElem			 BeforDraw	{ get; internal set;}	// 前の描画要素
		public seElem			 NextDraw	{ get; internal set;}	// 次の描画要素
		internal bool			 UpdateAnim { get; private set; }	// アニメーションが更新したか否か
		public evManager		 Event		{ get; set;			}	// イベント
		public seElemAnimManager Anim		{ get; private set; }	// アニメーション
		public seObjectWork		 ObjectWork { get; set;			}	// オブジェクトワーク
		public seElemState		 State		{ get; private set; }	// ステータスクラス

		// 子の数を取得
		private List<seElem> _lisChildList = new List<seElem>();
		public ReadOnlyCollection<seElem> Child { get { return this._lisChildList.AsReadOnly(); } }


		// 親と合成後のマトリクス取得
		public IGLMatrix NowMatrix { get { return this.State.NowMatrix; } }


		// 表示状態か否かを取得
		internal bool IsVisible { get { return (this.State.Visible != seAttrVisible.Hide) ? true : false; } }


		// コンストラクタ
		public seElem ( string name, seElem parentElem, int iIndex )
		{
			this.Hit = seHit.Non;
			this.Parent = parentElem;
			this.ElemID = iIndex;
			this.Name = name;

			if ( this.Parent != null )
			{
				this.Parent.AddChild( this );
			}

			this.State = new seElemState();		// データ管理
			this.Anim = new seElemAnimManager();	// アニメーション管理
		}

		// コンストラクタ(クローン)
		// 編集されないデータは共有してseElemのクローンを作成します
		private seElem( seElem mstElem, seElem parentElem, seSceneManager scene )
		{
			this.Hit = seHit.Non;
			this.Parent = parentElem;
			this.ElemID = mstElem.ElemID;
			this.Name = mstElem.Name;

			if ( this.Parent != null )
			{
				this.Parent.AddChild( this );
			}

			this.State = new seElemState();		// データ管理

			// 描画オブジェクトの再割り当て。
			seObjectWork w = mstElem.ObjectWork;
			this.ObjectWork = (w == null) ? null : w.Object.CreateDrawWork( this );

			// ステータスをコピー
			this.State.SetState( mstElem.State );
			// アニメーションをコピー
			this.Anim = mstElem.Anim;
			
			// イベントをコピー
			if ( mstElem.Event != null ) 
			{
				this.Event = mstElem.Event.Clone( this, scene );
			}
		}



		// 自分以下の全ての子に対して更新処理
		// @retval true:更新継続 false:下位の更新はしない
		internal bool Update ( IGLMatrix mCamera, float fFrameTime, float fTime )
		{
			// 自分以下のアニメーション全て停止中なら終了
			if ( this.StopAnimAll ) return false;
			// 自分がOFFなら終了
			if ( this.DisableEdit ) return false;

			// 自分のアニメーションだけ停止中なら
			if ( !this.StopAnim )
			{
				this.UpdateOne( mCamera, fFrameTime, fTime );
			}

			return true;
		}


		// アニメーションを任意のフレームの状態に設定。
		internal void ResetAnim(IGLMatrix mCamera, float fFrameTime, float fTime)
		{
			// 親のDisableEditParentか自分のDisableEditがtureの場合は無処理
			this.DisableEditParent = this.DisableEdit;
			if (this.Parent != null)
			{
				this.DisableEditParent |= this.Parent.DisableEditParent;
			}
			if (this.DisableEditParent) return;

			// ステータスアニメーションを開始前状態にする。
			this.State.RestartAnim();
			// 描画オブジェクトの内部ワークをリセット
			if (this.ObjectWork != null)
			{
				this.ObjectWork.Reset();
			}

			// アニメーションを更新(テスト)
			this.UpdateOne(mCamera, fFrameTime, fTime);
		}

		// 要素１つを更新
		private void UpdateOne( IGLMatrix mCamera, float fFrameTime, float fTime )
		{
			// アニメーションの値をステータスに適応
			this.UpdateAnim = this.State.UpdateAnim( fFrameTime, fTime );

			if ( this.Parent == null )
			{
				// rootアニメーションの時だけジャンプ判定をする
				seStateData data = this.State.GetData( emElemState.FrameCtrl );
				// キーを跨いだなら
				if ( data.MoveKey )
				{
					// フレームコントロールチェック
					((seStateFrameCtrl)data.Data).DoCheck();
				}
			}
			else
			{
				// 親がアニメーションしている場合はアニメーションする
				this.UpdateAnim |= this.Parent.UpdateAnim;
			}

			// 自分と親の継承関係を更新
			// 本来、ここが更新されると下位のエレメントにも影響を与えるので
			// ここ以下の全ての継承処理を更新する必要があります。
			// しかし、更新処理を一斉に行なう場合、無駄な継承処理が発生してしまう為
			// 下位の更新はしていません。
			this.Succession( mCamera );

			// 対象オブジェクトがあるなら状態を更新
			if ( this.ObjectWork != null )
			{
				this.ObjectWork.Update( fTime );
			}
		}

		// 親の要素のデータを継承して最終データを作成
		internal void Succession( IGLMatrix mCamera )
		{
			// 親がない(root)ならカメラマトリクスを掛ける
			seElem parent = this.Parent;
			if (parent == null)
			{
				// マトリクスのみ継承
				if ( mCamera != null )
				{
					this.State.Succession( mCamera );
				}
			}
			// 親があるなら継承
			else
			{
				// アニメーションを全て継承。
				this.State.Succession( parent.State );
			}
		}




		// データを複製し、クローンを作る
		// 再生時変更しないデータは複製せず同じものを参照します。
		public seElem Clone(seElem elemParent, seSceneManager scene)
		{
			seElem elem = new seElem(this, elemParent, scene);

			foreach (var childElem in this._lisChildList)
			{
				childElem.Clone(elem, scene);
			}
			return elem;
		}

		// 自分も含め、全ての子に処理を実行
		// funcの戻り値をfalseにすると途中終了
		public bool DoAll ( glbFunc<seElem, bool> func )
		{
			if (!func( this )) return false;

			foreach (seElem elem in this._lisChildList)
			{
				if ( !elem.DoAll( func ) ) return false;
			}
			return true;
		}


		// 子との関係を削除
		public void RemoveChild( seElem elem )
		{
			this._lisChildList.Remove( elem );
			elem.Parent = null;
		}

		// 子との関係を作成
		public void AddChild( seElem elem )
		{
			this._lisChildList.Add( elem );
			elem.Parent = this;
		}


		// データロード
		public void Load( IseStateInput input, seSceneManager scene )
		{
			this.ElemID = (int)input.ReadShort();
			seDefine.DebugTrace( "ElemID", this.ElemID );

			// あるなら要素名を入力
			this.Name = (scene.Option.IncludeElemName) ? input.ReadString()
																   : string.Empty;

			int iObjectID = (int)input.ReadShort();
			seObject obj = scene.ObjectManager.Find( iObjectID );
			this.ObjectWork = (obj==null) ? null : obj.CreateDrawWork(this);

			// 編集OFFフラグ
			this.DisableEdit = input.ReadBool();

			// 属性は保存しません。
			// イベントの途中状態を保存する等、必要に迫られたら考えます。
			// 

			// イベントを持っているならロード
			bool bHasEvent = input.ReadBool();

			// アニメーションロード
			this.Anim.Load( input, scene );

			if ( bHasEvent )
			{
				this.Event = new evManager( this, scene );
				this.Event.Load( input );
			}
			else
			{
				this.Event = null;
			}

			int iChildCount = (int)input.ReadShort();
			seDefine.DebugTrace( "ChildCount", iChildCount );
			for (int i = 0; i < iChildCount; i++)
			{
				seElem childElem = new seElem( string.Empty, this, seElemManager.InvalidID );
				childElem.Load( input, scene );
			}
		}

	}
}
