﻿using System;
using System.Diagnostics;
using glb;

namespace evc
{
	/**
	 * １要素状態管理クラス
	 * アニメーション管理から現在のデータをもらい、
	 * 継承後のデータを作成、管理します。
	 */
	public partial class seElemState
	{
		private seStateData[] asStateList;			// 継承データリスト
		private IGLMatrix _parentMatrixWork = null;	// 非継承フラグがある時に最終的に生成した親マトリクス
		internal glbVector2 BfoPos { get; private set; }	// 前の位置:主に直前との向きを算出する用

		// 表示フラグ取得(Show:表示, Hide:非表示, Succes:親を継承 の何れか)
		public seAttrVisible Visible { get; private set; }
		public IGLMatrix LocalMatrix { get; private set; }		// 親を考慮しないマトリクス
		public IGLMatrix ParentMatrix { get; private set; }		// 親マトリクスを取得
		public IGLMatrix NowMatrix { get; private set; }		// 親と合成後のマトリクス取得

		public glbColor NowColor { get; internal set; }
		public int NowPriority { get; private set; }			// 親と合成後の表示優先度を取得

		// エレメント共通ステータス
		public seStateInt1 Priority { get; private set; }		// 表示優先度
		public seStateAttribute  Attrbute { get; private set; }		// 属性
		public seStateColor		 Color	  { get; private set; }		// カラー
		public seStatePos		 Pos	  { get; private set; }		// 位置
		public seStateFloat2	 Scale	  { get; private set; }		// スケール
		public seStateFloat2	 Center	  { get; private set; }		// 中心位置
		public seStateRot		 Rot	  { get; private set; }		// 回転
		public seStateInt1		 Mask	  { get; private set; }		// マスクアルファ値


		// 全てのステータスをアニメーション開始前状態にする
		internal void RestartAnim ()
		{
			foreach (var data in this.asStateList)
			{
				if (data == null) continue;
				data.RestartAnim();
			}
			this.BfoPos = this.Pos.Value;
		}

		// アニメーションの更新
		internal bool UpdateAnim( float fFrameTime, float fTime )
		{
			this.BfoPos = this.Pos.Value;

			bool bUpdate = false;
			foreach (var data in this.asStateList)
			{
				if (data == null) continue;
				bUpdate |= data.Update( fFrameTime, fTime );
			}
			
			// アニメーションの更新があったならマトリクス等を再作成
			if ( bUpdate )
			{
				this.Apply();
			}
			return bUpdate;
		}

		internal void ClearAnim()
		{
			foreach (var data in this.asStateList)
			{
				if (data == null) continue;
				data.ChangeAnim(null);
			}
			this.BfoPos = this.Pos.Value;
		}



		// ステータス取得		
		public IseStateType this[ emElemState eType ]
		{
			get 
			{
				seStateData s = this.asStateList[ (int)eType ];
				return (s == null) ? null : s.Data;
			}
		}

		// データを取得
		public seStateData GetData(emElemState eType)
		{
			return this.asStateList[(int)eType];
		}


		// データを取得
		// クローンを作った方がいい？
		internal void SetState(seElemState state)
		{
			if (state == null) return;

			// 全ステータスを持つ配列を作成
			foreach (emElemState eType in EnumList<emElemState>.GetValues())
			{
				seStateData data = this.asStateList[(int)eType];
				if (data == null) continue;
				data.Data.Set(state[eType]);
			}
		}


		// 現在のデータから内部データを作成
		internal void Apply()
		{
			float fRot = (!seRootScene.Instance.EditMode && (this.Rot.SpinType == seSpinType.Direction))
						? seMath.GetRot(this.BfoPos, this.Pos.Value)
						: this.Rot.Rot;

			// 現在のデータからマトリクスを生成
			seMath.Transformation2D( this.LocalMatrix,
									this.Center.Value,
									this.Scale.Value,
									fRot,
									this.Pos.Value );
		}

		// マトリクスのみ継承
		internal void Succession( IGLMatrix mParent )
		{
			// 親マトリクスを保存
			this.ParentMatrix = mParent;

			this.NowMatrix.Copy( mParent );
			this.NowMatrix.Mul( this.LocalMatrix );
		}

		// データ継承
		internal void Succession( seElemState parentState )
		{
			// 継承しているか判定
			glbFunc<emElemState, bool> IsSuccession = delegate(emElemState eType)
			{
				return this.GetData( eType ).EnableSuccession;
			};

			// 描画順位は０なら親を継承
			int iNowPriority = this.Priority.Value;
			this.NowPriority = (iNowPriority == 0) ? parentState.NowPriority
												   : iNowPriority;

			// 今のカラーを取得
			glbColor c = this.Color.Color;
			// 継承するなら親のカラーと合成
			if ( IsSuccession( emElemState.Color) )
			{
				c.Mul( parentState.NowColor );
			}
			this.NowColor = c;

			// 表示の継承判定
			this.Visible = this.Attrbute.Visible;
			if ( this.Visible == seAttrVisible.Succes )
			{
				this.Visible = parentState.Visible;
			}

			bool bSuccesCenter = IsSuccession( emElemState.Center );
			bool bSuccesPos = IsSuccession( emElemState.Pos );
			bool bSuccesScale = IsSuccession( emElemState.Scale );
			bool bSuccesRot = IsSuccession( emElemState.Rot );

			// マトリクス要素を全て継承するなら
			if ( bSuccesCenter && bSuccesPos && bSuccesScale && bSuccesRot )
			{
				// 親のマトリクスをそのまま使う
				this.Succession( parentState.NowMatrix );
			}
			// マトリクス要素のどれかが非継承なら
			else
			{
				// 継承する要素のみでマトリクスを作る
				seMath.Transformation2D( this._parentMatrixWork,
					bSuccesCenter ? parentState.Center.Value : glbVector2.Zero, 
					bSuccesScale ? parentState.Scale.Value : glbVector2.One,
					bSuccesRot ? parentState.Rot.Rot : 0.0f, 
					bSuccesPos ? parentState.Pos.Value : glbVector2.Zero );
				this._parentMatrixWork.MulR( parentState.ParentMatrix );

				// 使用した親マトリクスに設定
				this.Succession( this._parentMatrixWork );
			}

		}


		// 描画前マテリアル設定
		internal void SetMaterial( siMaterial material )
		{
			// ブレンドモード設定:初期値はノーマル
			// ブレンドは親子間で継承しない。今のところ。
			BlendType eBlendType = BlendType.NORMAL;
			switch ( this.Attrbute.Blend )
			{
				case seAttrBlend.Add:	 eBlendType = BlendType.ADD;    break;
				case seAttrBlend.Sub:	 eBlendType = BlendType.NORMAL; break; // 減算は未実装の為
				case seAttrBlend.Normal:
				default:
					eBlendType = BlendType.NORMAL;
					break;
			}
			
			// マスクタイプの設定
			DoublyType eMaskType = DoublyType.Normal;
			bool bUpdateMaskAlpah = false;

			switch (this.Attrbute.Mask)
			{
				// マスク未使用
				case seAttrMask.NoUse:
					eMaskType = DoublyType.Normal;
					break;
				// マスクデータ
				case seAttrMask.Data:
					eMaskType = DoublyType.Mask;
					int iMaskAlpha = this.Mask.Value;
					if ( material.MaskAlpha != iMaskAlpha )
					{
						material.MaskAlpha = iMaskAlpha;
						bUpdateMaskAlpah = true;
					}
					break;
				// マスク適応
				case seAttrMask.Reflection:
					eMaskType = DoublyType.MaskReflection;
					break;
				// マスク適応＆終了
				case seAttrMask.End:
					eMaskType = DoublyType.MaskReflection;
					break;
				default:
					break;
			}

			if ((material.BlendType != eBlendType) || 
				(material.DoublyType != eMaskType) ||
				 bUpdateMaskAlpah)
			{
				material.BlendType = eBlendType;
				material.DoublyType = eMaskType;
				material.Apply();
			}
		}

		// スーテタスデータをリンク
		// 以降、設定されたデータにアニメーションが収められます。
		internal void LinkStateData( emElemState eState, IseStateType inStateType )
		{
			this.asStateList[ (int)eState ] = new seStateData( inStateType, true );
		}

		// 拡張した内部データを削除
		internal void ClearStateData()
		{
			for (int i = (int)seDefine.LastCommonState + 1; i < this.asStateList.Length; i++)
			{
				this.asStateList[ i ] = null;
			}
		}

		// コンストラクタ
		internal seElemState ()
		{
			this.Visible = seAttrVisible.Succes;
			this.NowColor = glbColor.White;
			this.NowPriority = 0;
			this.Priority = (seStateInt1)seStateUtil.Instance.CreateStateType( emElemState.Priority );
			this.Attrbute = (seStateAttribute)seStateUtil.Instance.CreateStateType( emElemState.Attribute);
			this.Color = (seStateColor)seStateUtil.Instance.CreateStateType( emElemState.Color);
			this.Pos = (seStatePos)seStateUtil.Instance.CreateStateType( emElemState.Pos);
			this.Scale = (seStateFloat2)seStateUtil.Instance.CreateStateType( emElemState.Scale );
			this.Center = (seStateFloat2)seStateUtil.Instance.CreateStateType( emElemState.Center );
			this.Rot = (seStateRot)seStateUtil.Instance.CreateStateType( emElemState.Rot );
			this.Mask = (seStateInt1)seStateUtil.Instance.CreateStateType( emElemState.MaskAlpha );

			// 全ステータスを持つ配列を作成
			emElemState[] inSateList = EnumList<emElemState>.GetValues();
			this.asStateList = new seStateData[ inSateList.Length ];

			// 共通ステータスを初期化
			this.asStateList[ (int)emElemState.Attribute ] = new seStateData( this.Attrbute, true );
			this.asStateList[ (int)emElemState.Priority ] = new seStateData( this.Priority, true );
			this.asStateList[ (int)emElemState.Pos ] = new seStateData( this.Pos, true );
			this.asStateList[ (int)emElemState.Center ] = new seStateData( this.Center, true );
			this.asStateList[ (int)emElemState.Scale ] = new seStateData( this.Scale, true );
			this.asStateList[ (int)emElemState.Rot ] = new seStateData( this.Rot, true );
			this.asStateList[ (int)emElemState.MaskAlpha ] = new seStateData( this.Mask, true );
			this.asStateList[ (int)emElemState.Color ] = new seStateData( this.Color, true );

			this.Attrbute.Mask = seAttrMask.NoUse;
			this.Attrbute.Visible = seAttrVisible.Succes;
			this.Attrbute.Blend = seAttrBlend.Normal;
		
			siIOManager inIOManager = seRootScene.Instance.IOManager;
			this.LocalMatrix = inIOManager.CreateMatrix();
			this.LocalMatrix.Identity();
			this.NowMatrix = inIOManager.CreateMatrix();
			this.NowMatrix.Identity();
			this._parentMatrixWork= inIOManager.CreateMatrix();
			this._parentMatrixWork.Identity();
		}
	}
}
