﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Diagnostics;

namespace evc
{
	// KeyIndex用特殊コード
	internal class KeyCode
	{
		internal const int First	= -1;	// 初めてのアニメーション
		internal const int End		= -2;	// 終了
	}

	// ステータス管理構造体
	public class seStateData
	{
		private IseStateType _beforData;				// １つ前

		// 再生メンバ
		private float fTime;						// 現在の経過時間
		private int	iKeyIndex = KeyCode.First;	// 最後に通過したキー
		private int	iNextFrame = 0;				// 次のキーまでのフレーム
		private seStateKeyManager _keyManager = null;		// 再生中のキーマネージャ

		// フラグ専用キー用コントロールメンバ
		private float fStateTime = 0.0f;			// データキーとデータキーの間の経過時間

		// フラグ専用キーリスト
		private List<seStateKeyInfo> _lisFlgKey = new List<seStateKeyInfo>();
		private seStateJump	_jumpStack = new seStateJump();

		public IseStateType Data { get; private set; }	// ステータスデータの取得・設定
		internal bool MoveKey { get; private set; }	// アニメーションの更新時、キーを越えたか否か
		public bool Visible { get; set; }			// 表示許可フラグ
		public bool EnableSuccession { get; set; }			// 継承許可フラグ


		// コンストラクタ
		internal seStateData ( IseStateType data, bool bEnable )
		{
			this.Data = data;
			this.Visible = true;
			this.EnableSuccession = bEnable;
			this._beforData = data.Clone();
		}


		// アニメーション開始前状態にする
		internal void RestartAnim()
		{
			this.fStateTime = 0.0f;
			this.fTime = 0.0f;
			this.iKeyIndex = KeyCode.First;
			this.iNextFrame = 0;
			// 現在のデータを１つ前のデータとする。
			this._beforData.Set( this.Data );
			
			// ループ等のキージャンプ用データのクリア
			this._lisFlgKey.Clear();
			this._jumpStack.Clear();
		}

		// アニメーションが変更された時の処理
		internal void ChangeAnim( seStateKeyManager inNewKeyManger )
		{
			// アニメーション開始前状態にする
			this.RestartAnim();
			// 新しいキーマネージャを設定
			this._keyManager = inNewKeyManger;
		}

		// アニメーションの更新
		// @retvla true:更新した false:更新しなかった
		internal bool Update( float fFrameTime, float fTime )
		{
			// 一度キー通過フラグをクリア
			this.MoveKey = false;
			
			if (this._keyManager == null) return false;

			// 初めてなら目標キーを設定
			if ( this.iKeyIndex == KeyCode.First )
			{
				this.MoveNextDataKey();
			}
			// 終了なら無処理
			else if (this.iKeyIndex == KeyCode.End)
			{
				return false;
			}

			ReadOnlyCollection<seStateKeyInfo> keys = this._keyManager.Keys;

			// キーが範囲外の場合(既にアニメーションが終了している等)は無処理
			if ((keys.Count <= this.iKeyIndex) || (this.iKeyIndex < 0))
				return false;

			seStateKeyInfo	nextKey = keys[ this.iKeyIndex ];
			float fKeyTime = 0.0f;

			// １フレームのミリ秒
			this.fTime += fTime;
			// データキー間の移動時間を加算
			this.fStateTime += fTime;

			// 今の時間の直前のキーを取得
			while (nextKey != null)
			{
				fKeyTime = this.iNextFrame * fFrameTime;

				// キーを通り過ぎていないなら終了
				if (this.fTime < fKeyTime) break;


				this.Data.Add(nextKey.Data);
				this._beforData.Add(nextKey.Data);

				// 次のキーへ移動
				this.fTime -= fKeyTime;
				this.fStateTime = this.fTime;

				// キー越えフラグをセット
				this.MoveKey = true;
				
				// データが待機なら次のキーに行かず終了
				if (nextKey.MoveControl == seKeyMoveContorl.Wait)
					return true;

				// 次のキーを取得
				nextKey = this.GetNextKey();
			}

			// キーの間なら
			if (nextKey != null)
			{
				// フラグキー管理
				if (this.UpdateFlagKey(fFrameTime))
				{
					nextKey = keys[this.iKeyIndex];
				}

				if (nextKey.HasSpeed)
				{
					float fRate = nextKey.GetRate(this.fTime / fKeyTime);
					this.Data.Anim(nextKey.Data, this._beforData, fRate);
				}
			}
			else
			{
				this.iKeyIndex = KeyCode.End;
			}
			
			return true;
		}


		// フラグ専用キー判定
		// @retval true:キーが移動した false:キーは移動しない
		private bool UpdateFlagKey( float fFrameTime )
		{
			while (this._lisFlgKey.Count > 0)
			{
				seStateKeyInfo key = this._lisFlgKey[0];
				float fNextFrame = fFrameTime * key.Frame;

				if (this.fStateTime < fNextFrame) break;

				// キー通過処理
				// キーのジャンプ(Loop)が発生した場合は
				// フラグ専用キーがクリアされているので即時に終了
				if (this.PassageKey(key)) return true;

				this.fStateTime -= fNextFrame;
				this._lisFlgKey.Remove(key);
			}

			return false;
		}


		// キーデータの内部状態をiStartIDを含めた
		// 次のデータのあるキーまで移動した状態にする。
		private bool MoveNextDataKey()
		{
			int iStartID = (this.iKeyIndex == KeyCode.First) ? 0 : this.iKeyIndex + 1;
			this.iNextFrame = 0;
			this._lisFlgKey.Clear();

			ReadOnlyCollection<seStateKeyInfo> keys = this._keyManager.Keys;
			for (int i = iStartID; i < keys.Count; i++)
			{
				// 次のデータのあるキーなら
				seStateKeyInfo nextKey = keys[ i ];
				this.iNextFrame += nextKey.Frame;

				// データがあるキーなら終了
				if ( !nextKey.DisableData )
				{
					this.iKeyIndex = i;
					return true;
				}
				// データなし(フラグ専用)キーなら保存
				else
				{
					this._lisFlgKey.Add( nextKey );
				}
			}
			return false;
		}


		// 現在のキーから次のキーを取得
		private seStateKeyInfo GetNextKey()
		{
			ReadOnlyCollection<seStateKeyInfo> keys = this._keyManager.Keys;

			// キー通過時処理
			if ( this.PassageKey(keys[ this.iKeyIndex ]) )
			{
				// 通過時にキーのジャンプがあったのでそのまま返す
				return keys[ this.iKeyIndex ];
			}

			// keyデータと検索開始位置からのフレーム数を設定
			if ( this.MoveNextDataKey() )
			{
				return keys[ this.iKeyIndex ];
			}

			return null;
		}


		// キー通過時処理
		// @retval true  : キーのジャンプがあった
		//		   false : キーのジャンプはない
		private bool PassageKey( seStateKeyInfo key )
		{
			bool bRet = false;

			// 編集時はループを無視
			if ( !seRootScene.Instance.EditMode )
			{
				// 通過したキーに対してループ判定をする。
				switch (key.MoveControl)
				{
					// ループ開始位置ならスタックに登録
					case seKeyMoveContorl.LoopStart:
						this._jumpStack.AddStack( key );
						break;
					// ループ終了位置なら
					case seKeyMoveContorl.LoopEnd:
						seStateKeyJumpStack stack = this._jumpStack.IncLoopNum();
						if ( stack != null ) 
						{
							// ループが指定回数終わったなら
							if ( ( key.LoopCount != seStateKeyInfo.Infinity ) && 
								 ( stack.iLoopNum > key.LoopCount ) )
							{
								// スタックを１つ減らす
								this._jumpStack.Pop();
							}
							// 指定回数ないなら
							else
							{
								// ループ開始位置に飛ぶ
								this.iKeyIndex = this._keyManager.Keys.IndexOf(stack.Key);
								// 次のキーまでのフレーム
								this.iNextFrame = key.ReturnFrame;
								// キーフレームをジャンプする時は色々なものを初期化する必要がある。
								this._lisFlgKey.Clear();
								bRet = true;
							}
						}
						break;
					default:
						break;
				}
			}

			// 継承するか否かを設定
			switch (key.Success)
			{
				case seKeySuccess.On: this.EnableSuccession = true; break;
				case seKeySuccess.Off: this.EnableSuccession = false; break;
				default: break;
			}

			return bRet;
		}


	}
}
