﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using glb;

namespace evc
{
	/// <summary>
	/// Description of seStateType.
	/// ステータスタイプインタフェース
	/// 四則演算・親子継承インタフェース
	/// </summary>
	public partial interface IseStateType
	{
		// データの設定
		// データを上書きします
		void Set(IseStateType state);

		// データの加算
		// 未設定の場合はム処理、それ以外のステータスを上書きします。
		void Add(IseStateType state);

		// アニメーション
		// @param next :次のデータ
		// @param befor :１つ前のデータ
		// @param fRate   :割合(0.0f=bfoData ～ 1.0f=this) 
		// 特殊なアニメーションはこれをオーバーライドして下さい。
		void Anim(IseStateType next, IseStateType before, float fRate );

		// 入力
		void Load(IseStateInput input);

		// 複製
		IseStateType Clone();
	}


	// ステータス作成拡張用:編集時のみ作成されるステータス
	interface ICreateStateEx
	{
		IseStateType CreateStateTypeEx( emElemState eStateType );
	}


	// ステータス作成
	public partial class seStateUtil
	{
		private static seStateUtil _instance = new seStateUtil();
		public static seStateUtil Instance { get { return seStateUtil._instance; } }

		// コンストラクタ
		private seStateUtil(){}


		// 最終データ保存用ステータスを作成
		public IseStateType CreateStateType( emElemState eStateType )
		{
			IseStateType state = null;
			switch ( eStateType )
			{
				// 先頭がジャンプ先。後は監視リソース
				case emElemState.FrameCtrl: state = new seStateFrameCtrl(); break;
				case emElemState.Attribute: state = new seStateAttribute(); break;
				// 回転
				case emElemState.Rot: state = new seStateRot( 0.0f, 0, seSpinType.Non ); break;

				// スケール
				case emElemState.Scale: state = new seStateFloat2( glbVector2.One ); break;

				case emElemState.Pos:			// 位置
					state = new seStatePos(); break;
				case emElemState.UV:			// UV
				case emElemState.Center:		// 中心位置
					state = new seStateFloat2( glbVector2.Zero ); break;

				case emElemState.Texture: state = new seStateSwitchInt1( 0 ); break;
				case emElemState.TextureControl: state = new seStateTextureControl(); break;
//				case emElemState.Priority: state = new seStateSwitchInt1( 0 ); break;
				case emElemState.Priority: state = new seStateInt1( 0 ); break;

				// ループ
				case emElemState.Loop: state = new seStateLoop( false, 1, false, 1 ); break;
				// カラー
				case emElemState.Color: state = new seStateColor(); break;

				// テキストコントロール
				case emElemState.TextControl : state = new seStateTextControl(); break;

				// ボリューム
				case emElemState.Volume: state = new seStateInt1( 1000 ); break;
				// マスクアルファ
				case emElemState.MaskAlpha: state = new seStateInt1( 255 ); break;
				// シーンアニメーション設定
				case emElemState.SceneAnim: state = new seStateString(); break;
				// シーンリソースロード設定
				case emElemState.Load: state = new seStateLoadControl(); break;
				// パン
				case emElemState.Pan: state = new seStateInt1( 0 ); break;
				// 数値
				case emElemState.Num: state = new seStateInt1( 0 ); break;
				// 数値ステータス
				case emElemState.NumState :
					seStateNumInfo l = new seStateNumInfo();
					l.MaxDisit = 5;
					state = l;
					break;
				case emElemState.SoundControl : state = new seStateSoundControl(); break;

				// グリッド制御点
				case emElemState.GridPos1:
				case emElemState.GridPos2:
				case emElemState.GridPos3:
				case emElemState.GridPos4:
					state = new seStateGridPoint(); break;

				// パーティクル情報
				case emElemState.ParticleControl:
					state = new seStateParticles(); break;
				default:
					ICreateStateEx creater = this as ICreateStateEx;
					state = creater.CreateStateTypeEx(eStateType);
					break;
			}

			Debug.Assert( state != null, "ステータスが作成できませんでした。");
			return state;
		}
	}



	// float2つ汎用クラス
	public partial class seStateFloat2 : IseStateType
	{
		public glbVector2 Value { get; set; }

		public seStateFloat2 () : this( glbVector2.Zero ){}
		public seStateFloat2 ( glbVector2 sInVector )
		{
			this.Value = sInVector;
		}

		// クローン
		public IseStateType Clone () { return new seStateFloat2( this.Value ); }

		public void Set(IseStateType state)
		{
			this.Value = ((seStateFloat2)state).Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			this.Value	= seIOUtil.ReadVector2( input );
		}

		public void Anim(IseStateType next, IseStateType before, float fRate)
		{
			glbVector2 b = ((seStateFloat2)before).Value;
			glbVector2 v = ((seStateFloat2)next).Value;
			v.Sub( b );
			v.Scale( fRate );
			v.Add( b );
			this.Value = v;
		}
	}

	// 汎用float1つ
	public partial class seStateFloat1 : IseStateType
	{
		public float Value { get; set; }

		public seStateFloat1() : this(0.0f){}
		public seStateFloat1( float fValue )
		{
			this.Value = fValue;
		}

		// クローン
		public IseStateType Clone () { return new seStateFloat1( this.Value ); }


		public void Set(IseStateType state)
		{
			this.Value = ((seStateFloat1)state).Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			this.Value = input.ReadFloat();
		}

		public void Anim(IseStateType next, IseStateType before, float fRate)
		{
			float fBefo = ((seStateFloat1)before).Value;
			float fNext = ((seStateFloat1)next).Value;
			this.Value = (fNext - fBefo) * fRate + fBefo;
		}
	}
	
	// int2つ汎用クラス
	public partial class seStateInt2 : IseStateType
	{
		public glbVector2I Value { get; set; }

		public seStateInt2() : this(glbVector2I.Zero){}
		public seStateInt2( glbVector2I sValue )
		{
			this.Value = sValue;
		}

		public IseStateType Clone ()
		{
			return new seStateInt2( this.Value );
		}

		public void Set(IseStateType state)
		{
			seStateInt2 si2 = (seStateInt2)state;
			this.Value = si2.Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			int iX = input.ReadInt();
			int iY = input.ReadInt();
			this.Value = new glbVector2I( iX, iY );
		}

		public void Anim(IseStateType next, IseStateType before, float fRate)
		{
			glbVector2I bi = ((seStateInt2)before).Value;
			glbVector2I ni = ((seStateInt2)next).Value;
			ni.Sub(bi);
			ni.X = (int)((float)ni.X * fRate);
			ni.Y = (int)((float)ni.Y * fRate);
			ni.Add(bi);
			this.Value = ni;
		}
	
	}

	// int1つ汎用クラス
	public partial class seStateInt1 : IseStateType
	{
		public int Value { get; set; }

		public seStateInt1(){}
		public seStateInt1( int iNum )
		{
			this.Value = iNum;
		}
		// クローン
		public IseStateType Clone () { return new seStateInt1( this.Value ); }

		public void Set(IseStateType state)
		{
			this.Value = ((seStateInt1)state).Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			this.Value = input.ReadInt();
		}

		public void Anim(IseStateType next, IseStateType before, float fRate)
		{
			int iBefo = ((seStateInt1)before).Value;
			int iNext = ((seStateInt1)next).Value;
			this.Value = (int)((float)(iNext - iBefo) * fRate) + iBefo;
		}
	}

	// アニメーションをしないint1汎用クラス
	public partial class seStateSwitchInt1 : IseStateType
	{
		public int Value { get; set; }

		// コンストラクタ
		public seStateSwitchInt1(){}
		public seStateSwitchInt1( int iNum )
		{
			this.Value = iNum;
		}

		public void Set(IseStateType state)
		{
			this.Value = ((seStateSwitchInt1)state).Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			this.Value = input.ReadInt();
		}

		// クローン
		public IseStateType Clone() { return new seStateSwitchInt1( this.Value ); }
		// アニメーションをしない
		public void Anim(IseStateType next, IseStateType before, float fRate){}
	}


	// 文字列クラス
	public partial class seStateString : IseStateType
	{
		public string Value { get; set; }


		public seStateString() : this(""){}
		public seStateString( string str )
		{
			this.Value = str;
		}
	
		// クローン
		public IseStateType Clone ()
		{
			return new seStateString( this.Value );
		}

		public void Set(IseStateType state)
		{
			this.Value = ((seStateString)state).Value;
		}

		public void Add(IseStateType state){ this.Set(state); }

		public void Load(IseStateInput input)
		{
			this.Value = input.ReadString();
		}

		public void Anim(IseStateType next, IseStateType before, float fRate){}
	}
	
	
}
