﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using Microsoft.DirectX.DirectSound;

namespace evcIO
{
	internal class WavFormatCtrl
	{
		private float		fStart;				// 再生開始タイム
		private float		fTime;				// 再生時間


		// WAVサイズ取得・設定
		private int iWavSize;
		public int WavSize { get { return this.iWavSize; } set{ this.iWavSize = value;} }

		private long lWavOffset;				// 自由に使える
		public long WavOffset { get { return lWavOffset; } set { lWavOffset = value; } }
	
		// 終了位置の有効を設定・取得
		// 終了位置が無効の場合は最後まで再生
		private bool bEndLimitEnable;	// 途中終了フラグ
		public bool EndLimit	{get { return bEndLimitEnable; } set { bEndLimitEnable = value; }}

		// 開始位置をバイト位置で取得
		private int iStartPos;			// ストリームの再生開始位置
		public int StartPos		{get { return iStartPos; }}

		// 終了位置をバイト位置で取得
		private int iEndPos;			// ストリームの再生終了位置
		public int EndPos		{get { return iEndPos; }}

		private WaveFormat	sWavFormat;
		public WaveFormat Format { get { return sWavFormat; }}

		// 波形の最後のデータ位置
		public long LastWavePos { get { return this.lWavOffset + this.iWavSize; } }


		// １サンプルのバイトが１byte以下の場合(例：4bit)があるので
		// フロートで表現
		public float SampleSize
		{
			get
			{
				// サンプルのバイト×チャンネル数
				return (float)this.sWavFormat.BitsPerSample / 8.0f * this.sWavFormat.Channels;
			}
		}
		// 再生開始時間
		public float Start
		{
			get { return fStart; }
			set
			{
				fStart = value;
				this.ResetPos();
			}
		}
		// 再生時間
		public float Time
		{
			get { return fTime; }
			set
			{
				fTime = value;
				this.ResetPos();
			}
		}





		// コンストラクタ
		public WavFormatCtrl ()
		{
			this.fStart = 0.0f;
			this.fTime = 0.0f;
			this.bEndLimitEnable = false;
		}

		// コンストラクタ(クローン用)
		private WavFormatCtrl( WavFormatCtrl inBase )
		{
			this.sWavFormat = inBase.sWavFormat;
			this.fStart		= inBase.fStart;			// 再生開始タイム
			this.fTime		= inBase.fTime;				// 再生時間
			this.lWavOffset		= inBase.lWavOffset;				// 自由に使える
			this.bEndLimitEnable = inBase.bEndLimitEnable;	// 途中終了フラグ
			this.iStartPos	= inBase.iStartPos;			// ストリームの再生開始位置
			this.iEndPos	= inBase.iEndPos;			// ストリームの再生終了位置
			this.iWavSize	= inBase.iWavSize;
		}

		// クローン作成
		public WavFormatCtrl Clone()
		{
			return new WavFormatCtrl( this );
		}



		// １秒が何バイトかを取得
		public int GetSecByte()
		{
			return (int)(this.SampleSize * this.sWavFormat.SamplesPerSecond);
		}

		// 開始、終了位置を再設定
		private void ResetPos()
		{
			float fSec = (float)this.GetSecByte();

			this.iStartPos = (int)(fSec * this.fStart);
			this.iStartPos -= (int)((float)this.iStartPos % this.SampleSize);
			
			this.iEndPos = (int)( fSec * (this.fStart + this.fTime) );
			this.iEndPos -= (int)((float)this.iEndPos % this.SampleSize);
		}


		// ストリームを指定の再生位置に移動(時間指定)
		public bool SetTime( Stream inStream, float fTime )
		{
			long lStart = (long)( fTime * this.GetSecByte() );
			// 開始位置はSampleSizeで割り切れる箇所からでないといけない。
			if ( lStart > 0 )
			{
				lStart -= lStart % (long)this.SampleSize;
			}

			inStream.Position = ( (long)this.WavOffset +	// ファイル内の波形の先頭位置
								  this.StartPos +	// 管理サウンドの先頭位置
								  lStart );				// 算出されたスタート位置
			return true;
		}
	
		
		// 登録
		internal void Register( Stream inStream )
		{
			inStream.Seek( 0, SeekOrigin.Begin );

			// wavフォーマットヘッダ
			BinaryReader inReader = new BinaryReader( inStream );
			inReader.ReadInt32();	// "RIFF"
			inReader.ReadInt32();	// size
			inReader.ReadInt32();	// WAVE

			// wavフォーマットInfo	
			ChunkInfo sChunkWork = new ChunkInfo();

			while (inReader.BaseStream.Length > inReader.BaseStream.Position)
			{
				sChunkWork.Register( inReader );

				switch ( sChunkWork.iChunkID )
				{
					// フォーマットなら
					case ChunkInfo.Fmt:
						WaveFormat sWavFormat = new WaveFormat();
						sWavFormat.FormatTag		= (WaveFormatTag)inReader.ReadInt16();
						sWavFormat.Channels			= inReader.ReadInt16();
						sWavFormat.SamplesPerSecond	= inReader.ReadInt32();
						sWavFormat.AverageBytesPerSecond = inReader.ReadInt32();
						sWavFormat.BlockAlign		= inReader.ReadInt16();
						sWavFormat.BitsPerSample	= inReader.ReadInt16();
						// 現状、DirectXではPcm以外サポートしていない。
						Debug.Assert( sWavFormat.FormatTag == WaveFormatTag.Pcm );
						// wavヘッダの登録
						this.sWavFormat = sWavFormat;

						break;
					// データなら
					case ChunkInfo.Data:
						// Tagには波形データ開始オフセットを入れる。
						this.WavOffset = (long)sChunkWork.lOffset;
						// 波形データサイズ
						this.WavSize = sChunkWork.iChunkSize;
						break;
					// それ以外
					default:
						break;
				}

				sChunkWork.NextChunk( inReader );
			}

			this.ResetPos();
	
			// ストリームをデータの先頭位置へ
			this.SetTime( inStream, 0.0f );
		}
	
	}
}
