﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Diagnostics;

namespace evcIO
{
	// メッシュコントロール
	// １．UVを色に変換したUVバッファを作成
	// ２．ピクセルシェーダでイメージに変換しながら描画

	// UVは0～65535で指定。
	// メッシュサイズの限界は特にありませんが、UVの精度の限界が65535の為
	// これ以上大きなテクスチャを描画しようとすると、イメージが端折られます。



	// クリップ判定
	[Flags]
	internal enum Clip : int
	{
		Non = 0,
		L	= 1,
		R	= 2,
		T	= 4,
		B	= 8,

		X = Clip.L | Clip.R,
		Y = Clip.B | Clip.T,
	}


	// 位置データ
	internal class PosInfo
	{
		public double x;
		public double y;
		public int tU;
		public int tV;
		public Clip clip;

		// 範囲外か判定
		public void CheckClip( Size size )
		{
			this.clip = Clip.Non;
			if (this.x <		    0) this.clip |= Clip.L;
			if (this.x >=  size.Width) this.clip |= Clip.R;
			if (this.y <		    0) this.clip |= Clip.T;
			if (this.y >= size.Height) this.clip |= Clip.B;
		}
	}

	// メッシュエフェクトタイプ
	internal enum MeshEffectType
	{
		Normal,
		Add,
	}


	internal class Mesh
	{
		private WriteableBitmap wb;		// ウェイトバッファ
		private BitmapImage bmpi;		// 描画イメージ

		// メッシュエフェクト；
		public UVMapEffectBase Effect { get; private set; }

		// 表示メッシュ
		public Image MeshImage { get; private set; }

		// 頂点データ
		public PosInfo[] Vertex { get; private set; }

		// 連結indexデータ
		public short[] Index { get; private set; }

		// メッシュ範囲
		public Size MeshSize { get; private set; }

		// エフェクトリスト
		private Dictionary<MeshEffectType,UVMapEffectBase> dicEffect;


		public void SetEffect( MeshEffectType type )
		{
			this.Effect = this.dicEffect[ type ];
			this.MeshImage.Effect = this.Effect;
		}


		// 連結indexデータ設定
		// index3つで１つの三角形を現します。
		public void SetIndex( short[] aiIndexList )
		{
			// メッシュのアルゴリズムによっては
			// 描画辺リストを作る

			this.Index = aiIndexList;
		}

		// イメージ設定
		public void SetImage( BitmapImage bmp )
		{
			this.bmpi = bmp;

			// 表示画像を登録
			ImageBrush ib = new ImageBrush();
			ib.ImageSource = this.bmpi;

			foreach (var ef in this.dicEffect.Values)
			{
				ef.Texture = ib;
			}
		}


		public void Resize(int iWidth, int iHeight)
		{
			// ウェイト画像を登録
			this.wb = new WriteableBitmap( iWidth, iHeight );
			ImageBrush ibw = (ImageBrush)this.Effect.Weight;
			ibw.ImageSource = this.wb;
 
			// 適当なソースを設定しておかないと、
			// ピクセルシェーダーより前に非表示判定されるっぽい。
			this.MeshImage.Source = this.wb;
			this.MeshImage.Width = iWidth;
			this.MeshImage.Height = iHeight;
			// メッシュサイズ
			this.MeshSize = new Size((double)iWidth, (double)iHeight);
		}


		public Mesh( int iVertexCount, int iWidth, int iHeight )
		{
			// 頂点リスト作成
			this.Vertex = new PosInfo[ iVertexCount ];
			for (int i = 0; i < iVertexCount; i++)
			{
				this.Vertex[ i ] = new PosInfo();
			}

			// エフェクト作成
			ImageBrush ib = new ImageBrush();
			this.dicEffect = new Dictionary<MeshEffectType,UVMapEffectBase>()
			{
				{ MeshEffectType.Normal, new UVMapEffect(ib) },
				{ MeshEffectType.Add, new UVMapEffectAdd(ib) },
			};

			this.MeshImage = new Image();
			this.SetEffect(MeshEffectType.Normal);

			// ウェイト画像バッファを作成
			this.Resize( iWidth, iHeight );
		}


		// 水平ラインを描画
		private void DrawaLine( PoslineMove mc1, PoslineMove mc2 )
		{
			if ( mc1.NowX > mc2.NowX )
			{
				PoslineMove tm = mc1;
				mc1 = mc2;
				mc2 = tm;
			}

			int iX1 = (int)mc1.NowX;
			int u1 = (int)mc1.NowU;
			int v1 = (int)mc1.NowV;

			int istart = mc1.NowY * this.wb.PixelWidth + iX1;
			int ileng = (int)mc2.NowX - iX1 + 1;

#if true
			int tu = (int)mc2.NowU - u1;
			int tv = (int)mc2.NowV - v1;
			int u = 0;
			int v = 0;
			for (int i = 0; i < ileng; i++)
			{
				wb.Pixels[ i + istart ] = (((u / ileng) + u1) & 0xffff) |
										  ((((v / ileng) + v1) & 0xffff) << 16);
				u += tu;
				v += tv;
			}
#else
			// ワイヤーフレーム
			// 0x00ff = 255
			wb.Pixels[ istart ] = (u1 & 0xffff) | ((v1&0xffff) << 16);
			if ( ileng > 0 )
			{
				wb.Pixels[ istart + ileng ] = ((int)mc2.NowU & 0xffff) |
											  (((int)mc2.NowV & 0xffff) << 16);
			}
#endif	
		}


		// 頂点vをsize内に収まるようにクリップ
		private Point ClipPoint( Point v, Point v2, Size size, Clip clip )
		{
			if ( (clip & Clip.X) != Clip.Non )
			{
				double r =  (v2.Y - v.Y) / (v2.X - v.X);
				double dX = ((clip & Clip.L) != Clip.Non) ? 0.0 : size.Width - 1;
				v = new Point( dX, (r * (dX - v.X)) + v.Y);

				// クリップフラグを更新
				// Yしか使わないのでYのみを処理
				clip = Clip.Non;
				if ( v.Y <			  0 ) clip |= Clip.T;
				if ( v.Y >= size.Height ) clip |= Clip.B;
			}


			if ( (clip & Clip.Y) != Clip.Non )
			{
				double r  = (v2.X - v.X) / (v2.Y - v.Y);
				double dY = ((clip & Clip.T) != Clip.Non) ? 0.0 : size.Height - 1;
				v = new Point((r * (dY - v.Y)) + v.X, dY );

				// 計算の結果、Xが画面外に出ることがある。
				if ( v.X < 0.0 ) v.X = 0.0;
				if ( v.X >= size.Width ) v.X = size.Width - 1;
			}

			return v;
		}


		// ２点の距離を求める
		private double GetDistance ( Point p1, Point p2 )
		{
			double dx = p1.X - p2.X;
			double dy = p1.Y - p2.Y;
			return Math.Sqrt( dx * dx + dy * dy );
		}

		// 新しいPosInfoを作成
		private PosInfo CreateNewPos( PosInfo p, PosInfo p2, Point pos, double dRate )
		{
			PosInfo pp = new PosInfo();
			pp.x = pos.X;
			pp.y = pos.Y;
			pp.tU = (int)((double)(p.tU - p2.tU) * dRate) + p2.tU;
			pp.tV = (int)((double)(p.tV - p2.tV) * dRate) + p2.tV;
			return pp;
		}


		// 三角形を描画
		private PoslineMove mc1 = new PoslineMove();
		private PoslineMove mc2 = new PoslineMove();
		private List<PosInfo> lisPoly = new List<PosInfo>();
		private void DrawTriangle( PosInfo[/*3*/] lisPos )
		{
			Size size = this.MeshSize;
			this.lisPoly.Clear();

			int iTopPosID = 0;	// 多角形の一番上の頂点
			int iTopPosY = 0;	// 一番上のY
			bool bHLine = true;	// 全てが一直線のラインの場合


			// 登録
			Action<PosInfo> AddPoly = delegate( PosInfo pos )
			{
				if ( lisPoly.Count < 1 )
				{
					lisPoly.Add( pos );
					iTopPosY = (int)pos.y;
				}
				else
				{
					// 最後に登録したものと極端に近い場合は無処理
					PosInfo lastInfo = lisPoly[ lisPoly.Count - 1];
					double dx = lastInfo.x - pos.x;
					double dy = lastInfo.y - pos.y;
					// １ドット未満なら無処理
					if ( Math.Sqrt( dx * dx + dy * dy ) > 1.0 )
					{
						lisPoly.Add( pos );
						int iY = (int)pos.y;

						// 1ドット以上高さがある
						if (iY != iTopPosY) bHLine = false;

						if ( iY < iTopPosY )
						{
							iTopPosY = iY;
							iTopPosID = lisPoly.Count - 1;
						}

					}
				}
			};

			Point p_next = new Point();
			Point p = new Point();
			for (int i = 0; i < 3; i++)
			{
				PosInfo nextInfo = lisPos[ (i + 1)%3 ];
				p_next.X = nextInfo.x;
				p_next.Y = nextInfo.y;
				Clip clipNext = nextInfo.clip;

				PosInfo nowInfo = lisPos[ i ];
				p.X = nowInfo.x;
				p.Y = nowInfo.y;
				Clip clip = nowInfo.clip;

				double dMstLen = this.GetDistance( p, p_next );
				double dNewLen;
				PosInfo newPos;

				// クリップの必要がない場合
				if ( clip == Clip.Non )
				{
					// 元の頂点情報をそのまま入れる
					newPos = nowInfo;
					dNewLen = dMstLen;
				}
				// クリップする必要がある場合
				else
				{
					// 画面外の斜めに頂点があるなら、角に頂点を追加
					if (((clip & Clip.X) != Clip.Non) &&
						((clip & Clip.Y) != Clip.Non))
					{
						Point kp = new Point();
						kp.X = ((clip & Clip.L) != Clip.Non) ? 0.0 : (size.Width - 1);
						kp.Y = ((clip & Clip.T) != Clip.Non) ? 0.0 : (size.Height - 1);

						double dkLen = this.GetDistance(kp, p_next);

						// 新しいPosInfoを作成
						AddPoly( this.CreateNewPos(
							nowInfo, nextInfo, kp, dkLen / dMstLen));
					}

					// ２点が同じ境界で画面外：点は存在しない
					if ((clip & clipNext) != Clip.Non) continue;

					// クリップ
					p = this.ClipPoint( p, p_next, size, clip );

					dNewLen = this.GetDistance( p, p_next );
					// 新しいPosInfoを作成
					newPos = this.CreateNewPos(
							nowInfo, nextInfo, p, dNewLen/dMstLen );
				}
				AddPoly( newPos );


				// 次の点が範囲内の場合は無処理
				if ( clipNext != Clip.Non )
				{
					dMstLen = dNewLen;

					// クリップ
					p_next = this.ClipPoint(p_next, p, size, clipNext );
					dNewLen = this.GetDistance( p, p_next );
					// １ドット以下の場合はまとめる。
					if ( dNewLen < 1.0 ) continue;

					// クリップがあった場合は終了点も登録
					// 新しいPosInfoを作成
					AddPoly( this.CreateNewPos(
								nextInfo,
								newPos,
								p_next,
								dNewLen / dMstLen ));
				}
			}

			// 有効な点が２つ以下の場合は非表示
			if ( lisPoly.Count < 3 )
			{
				return;
			}


			// ３点の高さが同じの水平ラインの場合
			if ( bHLine )
			{
				PosInfo minPos = lisPoly[ 0 ];
				PosInfo maxPos = minPos;

				for (int i = 1; i < lisPoly.Count; i++)
				{
					PosInfo pos = lisPoly[ i ];
					if ( pos.x < minPos.x ) minPos = pos;
					if ( pos.x > maxPos.x ) maxPos = pos;
				}

				int iX1 = (int)minPos.x;
				int u1 = (int)minPos.tU;
				int v1 = (int)minPos.tV;

				int istart = iTopPosY * this.wb.PixelWidth + iX1;
				int ileng = (int)maxPos.x - iX1 + 1;

				int tu = (int)maxPos.tU - u1;
				int tv = (int)maxPos.tV - v1;
				int u = 0;
				int v = 0;
				for (int i = 0; i < ileng; i++)
				{
					this.wb.Pixels[ i + istart ] = (((u / ileng) + u1) & 0xffff) |
											  ((((v / ileng) + v1) & 0xffff) << 16);
					u += tu;
					v += tv;
				}
			}
			// 三角形描画
			else
			{
				mc1.Init( lisPoly, iTopPosID, 1 );
				mc2.Init( lisPoly, iTopPosID, -1 );

				while( !(mc1.IsEnded && mc2.IsEnded) )
				{
					this.DrawaLine( mc1, mc2 );
					mc1.MoveNextLine();
					mc2.MoveNextLine();
				}
			}
		}

		// ライン移動管理
		internal class PoslineMove
		{
			private List<PosInfo> Pos;	// ポイントリスト
			private PosInfo NowPos;		// 現在のポイント
			private int NextID;			// 次の位置のID

			// 本来は以下のステータスは{get; private set;}
			// にすべきですが、高速化を期待してメンバにしてみます。
			public double MoveX;	// Yが１下がる毎に加算されるX
			public double MoveU;	// Yが１下がる毎に加算されるU
			public double MoveV;	// Yが１下がる毎に加算されるV
			public double NowX;		// 現在のX
			public double NowU;		// 現在のU
			public double NowV;		// 現在のV
			public int NowY;		// 現在のY

			public bool IsEnded { get; private set; }	// 終了フラグ

			private int iMoveNum;	// ポイントの制御方向

			// 次の位置を取得
			public PosInfo NextPos { get { return this.Pos[ this.NextID ]; } }


			// 初期化
			public void Init( List<PosInfo> lisPos, int iTopID, int iMovePower )
			{
				this.iMoveNum = iMovePower;
				this.Pos = lisPos;
				this.NowY = (int)lisPos[ iTopID ].y;
				this.IsEnded = false;

				// ラインポイントを更新(ポイント直指定)
				this.UpdatePos( lisPos[ iTopID ], iTopID + this.iMoveNum );
			}

			// ラインポイントを更新(ポイント直指定)
			private bool UpdatePos( PosInfo nowPos, int nextID )
			{
				if ( this.IsEnded ) return false;

				if ( nextID >= this.Pos.Count )
				{
					nextID = 0;
				}
				else if ( nextID < 0 )
				{
					nextID = this.Pos.Count - 1;
				}


				this.NowPos = nowPos;
				this.NextID = nextID;

				int sX = (int)this.NowPos.x;
				int sY = (int)this.NowPos.y;
				int eX = (int)this.NextPos.x;
				int eY = (int)this.NextPos.y;
				this.NowY = sY;

				int dYLeng = eY - sY;
				// Yがマイナス：次の点が上向きになった。
				if ( dYLeng < 0 )
				{
					this.IsEnded = true;
					return false;
				}
				// 水平ラインの場合は更に次のポイントへ
				else if ( dYLeng == 0 )
				{
					return this.UpdatePos();
				}
				else
				{
					this.NowX = (double)sX;
					this.NowU = (double)this.NowPos.tU;
					this.NowV = (double)this.NowPos.tV;

					// 移動量更新
					this.MoveX = (double)(eX - sX) / (double)dYLeng;
					this.MoveU = (double)(this.NextPos.tU - this.NowU) / (double)dYLeng;
					this.MoveV = (double)(this.NextPos.tV - this.NowV) / (double)dYLeng;
				}

				return true;
			}
			
			// ラインポイントを更新
			private bool UpdatePos()
			{
				return this.UpdatePos( this.NextPos, this.NextID + this.iMoveNum );
			}

			// 次のラインへ移動
			public void MoveNextLine()
			{
				if ( this.IsEnded ) return;
				
				this.NowY ++;
				if ( this.NowY >= (int)this.NextPos.y )
				{
					this.UpdatePos();
				}
				else
				{
					this.NowX += this.MoveX;
					this.NowU += this.MoveU;
					this.NowV += this.MoveV;
				}
			}
		}


		// 状態を反映
		public void Apply()
		{
			if ( (this.Vertex == null) || (this.Index == null) ) return;

			PosInfo[] lisPos = new PosInfo[ 3 ];
			PosInfo[] vertex = this.Vertex;
			short[] index = this.Index;

			// 先にバッファをクリア
			// バッファは全て埋められるのでクリアしない方が良い？
			Array.Clear( this.wb.Pixels, 0, this.wb.Pixels.Length );
			for (int i = 0; i < this.Index.Length; i+= 3)
			{
				lisPos[0] = vertex[ index[i  ] ];
				lisPos[1] = vertex[ index[i+1] ];
				lisPos[2] = vertex[ index[i+2] ];
				this.DrawTriangle( lisPos );
			}
			this.wb.Invalidate();
		}
	}
}
