﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Diagnostics;
using System.Windows.Resources;
using System.IO;

namespace sl
{
	// ロード要求クラスインタフェース
	internal interface ISlLoadUser
	{

		// ロード終了時
		void Completed( SlLoadInfo inLoadInfo );
	}
	
	
	// １ロード管理
	internal class SlLoadInfo
	{
		internal bool				IsEnable { get; set; }		// 使用可能なデータか否かを取得
		internal StreamResourceInfo Stream	 { get; set; }		// ファイルリソースを取得
		internal string				Path	 { get; set; }		// ロードファイルパス

		// 使用可能になったとき呼ばれる関数
		private List<ISlLoadUser> lisFileInfo = new List<ISlLoadUser>();


		public SlLoadInfo()
		{
			this.Path = string.Empty;
		}
	
		// ロード要求をしたクラスを登録：データロード完了時にそれらのクラスに通知
		internal void Register( ISlLoadUser inLoadInfo )
		{
			// 既に登録済みなら無処理
			if ( this.lisFileInfo.Contains( inLoadInfo ) ) return;
			this.lisFileInfo.Add( inLoadInfo );
		}
	
		// ロード終了を通知
		internal void SendCompleted()
		{
			// 保持しているリストにロード終了を通知
			this.lisFileInfo.ForEach( delegate( ISlLoadUser inUser )
			{
				inUser.Completed( this );
			});

			this.lisFileInfo.Clear();
		}
	}
	
	
	// ロード管理
	// 後に参照カウンタつける予定。
	// 優先順位がつけられればより親切。
	//

	internal class SlLoadManager : IDisposable
	{
		// シングルトン
		private static SlLoadManager inInstance = new SlLoadManager();
		public static SlLoadManager Instance { get { return SlLoadManager.inInstance; } }

	
		private WebClient				inLoader = new WebClient();
//		private ISlLoading				inSendFunc = null;		// ロード状態通知用

		// 現在ロード中のファイルパス
		private string strLoadingFilePath = string.Empty;

		// ファイルストリームリスト
		private Dictionary<string, SlLoadInfo>	dicStream = new Dictionary<string,SlLoadInfo>();

		// ロード中フラグ
		public bool IsBusy	{ get; private set; }

		// 任意のファイルを取得
		public SlLoadInfo Find( string strFile )
		{
			if ( !this.dicStream.ContainsKey( strFile ) ) return null;
			return this.dicStream[ strFile ];
		}

		// コンストラクタ
		public SlLoadManager()
		{
			this.inLoader.OpenReadCompleted += new OpenReadCompletedEventHandler(inLoader_OpenReadCompleted);
//			this.inLoader.DownloadProgressChanged += new DownloadProgressChangedEventHandler(inLoader_DownloadProgressChanged);
		}


		// ロード状態が更新したなら
		void inLoader_DownloadProgressChanged(object sender, DownloadProgressChangedEventArgs e)
		{

			// 進捗を何処に通知するか考える(ファイルorシーンor全体)
#if false
			if ( this.inSendFunc != null )
			{
				// ロード済みのファイル数を数える
				int iLoadedCount = 0;
				foreach (SlLoadInfo inLoadInfo in this.dicStream.Values)
				{
					if ( inLoadInfo.IsEnable ) iLoadedCount++;
				}

#if false	// ファイルを使用するクラスに対してロード状態を通知するならこの辺を追記する。
				string strFullPath = e.Address.AbsolutePath;
				SlLoadInfo inActiveInfo = this.dicStream[strFullPath];
#endif
				this.inSendFunc.ChangeProgress( e.ProgressPercentage,
										iLoadedCount + 1, 
										this.dicStream.Count );
			}
 
#endif	
		}


		// ロード終了なら
		void inLoader_OpenReadCompleted(object sender, OpenReadCompletedEventArgs e)
		{
			string strPath = System.IO.Path.GetFileName( this.strLoadingFilePath );
			SlLoadInfo inLoadInfo = this.dicStream[strPath];

			inLoadInfo.Stream = new StreamResourceInfo(e.Result, null);
			inLoadInfo.IsEnable = true;
			inLoadInfo.SendCompleted();

			// ロード待ちのファイルがあればロード実行
			foreach (KeyValuePair<string, SlLoadInfo> inPair in this.dicStream)
			{
				// 既にロード済み
				if (inPair.Value.IsEnable) continue;

				// 次のファイルをロード	
				this.strLoadingFilePath  = inPair.Key;
				this.inLoader.OpenReadAsync(new Uri(inPair.Key, UriKind.Relative));
				return;
			}

			// 全てのファイルロード終了
			this.IsBusy = false;
			this.strLoadingFilePath = string.Empty;
		}


		// ロード開始
		public bool Load( string strFile, ISlLoadUser inUser )
		{
			if ( strFile == null ) return false;
			if ( strFile.Length == 0 ) return false;

			// 初めてロード要求があった場合
			SlLoadInfo inLoadInfo = null;
			if ( !this.dicStream.ContainsKey( strFile ) )
			{
				inLoadInfo = new SlLoadInfo();
				inLoadInfo.Path = strFile;
				this.dicStream[ strFile ] = inLoadInfo;
			}

			inLoadInfo = this.dicStream[ strFile ];

			// 既に使用可能状態なら
			if ( inLoadInfo.IsEnable )
			{
				inUser.Completed( inLoadInfo );
				return true;
			}

			// 未初期化リストに登録
			inLoadInfo.Register( inUser );

			// 何か別ファイルをロード中なら終了
			if ( this.IsBusy ) return true;
			this.IsBusy = true;

			// ロード開始
			this.strLoadingFilePath = strFile;

			this.inLoader.OpenReadAsync( new Uri( strFile, UriKind.Relative) );
			Debug.WriteLine( "Pack Load Start " + strFile );
			
			return true;
		}

		#region IDisposable メンバ

		// ファイルの解放
		public void Dispose()
		{
			this.inLoader.CancelAsync();
	
			foreach (SlLoadInfo inInfo in this.dicStream.Values)
			{
				// 既にロード済みなら解放
				if ( inInfo.IsEnable )
				{
					inInfo.Stream.Stream.Dispose();
				}
			}
		}

		#endregion
	}
}
